/*
 *   Copyright 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spago4Q.

    Spago4Q is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published 
    by the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spago4Q is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package it.eng.spago.util;

import it.eng.spago.base.Constants;
import it.eng.spago.base.RequestContainer;
import it.eng.spago.base.ResponseContainer;
import it.eng.spago.base.SourceBean;
import it.eng.spago.dbaccess.DataConnectionManager;
import it.eng.spago.dbaccess.SQLStatements;
import it.eng.spago.dbaccess.Utils;
import it.eng.spago.dbaccess.sql.DataConnection;
import it.eng.spago.dbaccess.sql.SQLCommand;
import it.eng.spago.dbaccess.sql.result.DataResult;
import it.eng.spago.dbaccess.sql.result.InformationDataResult;
import it.eng.spago.dbaccess.sql.result.ScrollableDataResult;
import it.eng.spago.error.EMFErrorSeverity;
import it.eng.spago.error.EMFInternalError;
import it.eng.spago.tracing.TracerSingleton;

import java.sql.Types;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * DATE CONTRIBUTOR/DEVELOPER NOTE 
 * 13-12-2004 Butano - SourceBean ora restituisce ArrayList
 * 13-12-2005 Boccalon/Grassi - Added utility methods to execute queries giving the DataConnection
 *            instead of the connection pool, or giving the parameter list instead of the
 *            automatic parameters retrieval
 * 9-1-2008 Gavardi: Modifid with double option (only for spago4Q)
 * 
 */

public abstract class QueryExecutor {
    public static final String CREATE = "CREATE";

    public static final String READ = "READ";

    public static final String UPDATE = "UPDATE";

    public static final String DELETE = "DELETE";

    /**
     * Crea il comando da eseguire a seconda del tipo di richiesta
     * 
     * @param dataConnection
     * @param query
     * @param type Il tipo di query da eseguire: CREATE, READ, UPDATE, DELETE
     * @return
     */
    public static SQLCommand createStatementSql(final DataConnection dataConnection,
            final String statement, final String type) {
        SQLCommand sqlCommand = null;
        if (type.equalsIgnoreCase("INSERT"))
            sqlCommand = dataConnection.createInsertCommand(statement);
        else if (type.equalsIgnoreCase("UPDATE"))
            sqlCommand = dataConnection.createUpdateCommand(statement);
        else if (type.equalsIgnoreCase("DELETE"))
            sqlCommand = dataConnection.createDeleteCommand(statement);
        else
            sqlCommand = dataConnection.createSelectCommand(statement);
        return sqlCommand;
    }

    /**
     * Apre la connessione attraverso il pool
     * 
     * @param pool
     * @return
     * @throws EMFInternalError
     */
    public static DataConnection openConnection(final String pool) throws EMFInternalError {
        return DataConnectionManager.getInstance().getConnection(pool);
    }

    /**
     * Esegue i comandi SQL SELECT, INSERT, DELETE, UPDATE. La connessione viene chiesta
     * al connection pool.
     * 
     * @param requestContainer
     * @param responseContainer
     * @param pool Il nome del pool del database
     * @param query Il SourceBean contenente la configurazione della query da eseguire
     * @param type Il tipo di query da eseguire: CREATE, READ, UPDATE, DELETE
     * @return
     */
    public static Object executeQuery(final RequestContainer requestContainer,
            final ResponseContainer responseContainer, final String pool, final SourceBean query,
            final String type) {
        Object result = null;
        DataConnection dataConnection = null;

        try {
            // Apre la connessione
            dataConnection = openConnection(pool);
            result = executeQuery(requestContainer, responseContainer, dataConnection, query, type);
        } // try
        catch (EMFInternalError ex) {
            TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.CRITICAL,
                    "QueryExecutor::executeQuery:", (Exception) ex);
            responseContainer.getErrorHandler().addError(
                    new EMFInternalError(EMFErrorSeverity.ERROR, ex.getNativeException()));
        } // catch (Exception ex) try
        finally {
            Utils.releaseResources(dataConnection, null, null);
        } // finally try
        return result;
    } // public static Object executeQuery(RequestContainer

    /**
     * Esegue i comandi SQL SELECT, INSERT, DELETE, UPDATE. La connessione 
     * passata come parametro (per le transazioni gestite manualmente)
     * 
     * @param requestContainer
     * @param responseContainer
     * @param dataConnection La connessione verso il database
     * @param query Il SourceBean contenente la configurazione della query da eseguire
     * @param type Il tipo di query da eseguire: CREATE, READ, UPDATE, DELETE
     * @return
     */
    public static Object executeQuery(final RequestContainer requestContainer,
            final ResponseContainer responseContainer, DataConnection dataConnection,
            final SourceBean query, final String type) {
        Object result = null;

        try {
            // Crea il comando da eseguire a seconda del tipo di richiesta
            String partialStatement = SQLStatements.getStatement((String) query
                    .getAttribute("STATEMENT"));
            ArrayList inputParameters = new ArrayList();
            StringBuffer statement = new StringBuffer(partialStatement);

            // Elaboro i parametri normali, ossia quelli di tipo "PARAMETER"
            List parameters = query.getAttributeAsList("PARAMETER");
            for (int i = 0; i < parameters.size(); i++) {
                SourceBean parameter = (SourceBean) parameters.get(i);
                handleParameter(requestContainer, responseContainer, parameter, inputParameters,
                        dataConnection);
            } // for

            // Elaboro i parametri di filtro, ossia quelli di tipo
            // "FILTER_PARAMETER"
            List filterParameters = query.getAttributeAsList("FILTER_PARAMETER");
            String condizioneSql = (partialStatement.indexOf("WHERE") != -1) ? " AND " : " WHERE ";
            for (int i = 0; i < filterParameters.size(); i++) {
                SourceBean filterParameter = (SourceBean) filterParameters.get(i);

                if (handleFilterParameter(requestContainer, responseContainer, filterParameter,
                        inputParameters, dataConnection, statement, condizioneSql)) {
                    condizioneSql = " AND ";
                }
            } // for

            // Elabora tutte le sezioni di tipo "GROUP_BY"
            SourceBean groupByParameters = (SourceBean) query.getAttribute("GROUP_BY");
            if (groupByParameters != null) {
                handleOrderByParameter(groupByParameters, statement, " GROUP BY ");
            }

            // Elabora tutte le sezioni di tipo "ORDER_BY"
            SourceBean orderByParameters = (SourceBean) query.getAttribute("ORDER_BY");
            if (orderByParameters != null) {
                handleOrderByParameter(orderByParameters, statement, " ORDER BY ");
            }
            
            result = executeQuery(dataConnection, statement.toString(), type, inputParameters);
        } // try
        catch (EMFInternalError ex) {
            TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.CRITICAL,
                    "QueryExecutor::executeQuery:", (Exception) ex);
            responseContainer.getErrorHandler().addError(
                    new EMFInternalError(EMFErrorSeverity.ERROR, ex.getNativeException()));
        } // catch (Exception ex) try
        catch (Exception ex) {
            TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.CRITICAL,
                    "QueryExecutor::executeQuery:", (Exception) ex);
            responseContainer.getErrorHandler().addError(
                    new EMFInternalError(EMFErrorSeverity.ERROR, ex));
        } // catch (Exception ex) try
        return result;
    } // public static Object executeQuery(RequestContainer

    /**
     * Esegue statement SQL con parametri passati esplicitamente nel metodo
     * 
     * @param dataConnection La connessione di database: non viene effettuata alcuna commit
     * o rollback
     * @param type Il tipo di query da eseguire: CREATE, READ, UPDATE, DELETE
     * @param query Il SourceBean contenente la configurazione della query da eseguire
     * @param parameters I parametri da integrare nello statement
     * @return
     * @throws Exception
     */
    public static Object executeQuery(DataConnection dataConnection, String type, SourceBean query,
            ArrayList parameters) throws Exception {
        TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.DEBUG,
                "QueryExecutor::executeQuery: invocato");
        Object result = null;
        String statement = SQLStatements.getStatement((String) query.getAttribute("STATEMENT"));
        try {
            ArrayList inputParameters = new ArrayList();
            if (parameters != null) {
                for (Iterator it = parameters.iterator(); it.hasNext();) {
                    inputParameters.add(dataConnection.createDataField("", Types.VARCHAR,
                            (String) it.next()));
                }
            }
            result = executeQuery(dataConnection, statement, type, inputParameters);
            return result;
        } // try
        catch (Exception ex) {
            TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.CRITICAL,
                    "QueryExecutor::executeQuery:", ex);
            throw ex;
        } // catch (Exception ex) try
    } // public static SourceBean executeQuery(DataConnection dataConnection,

    protected static Object executeQuery(DataConnection dataConnection, final String statement,
            final String type, final ArrayList inputParameters) throws Exception {
        TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.DEBUG,
                "QueryExecutor::executeQuery: invocato");
        SQLCommand sqlCommand = null;
        DataResult dataResult = null;
        Object result = null;
        try {
            // Crea il comando da eseguire a seconda del tipo di richiesta
            sqlCommand = createStatementSql(dataConnection, statement, type);
            if ((inputParameters != null) && (inputParameters.size() != 0)) {
                dataResult = sqlCommand.execute(inputParameters);
            } else
                dataResult = sqlCommand.execute();
            if (type.equalsIgnoreCase("SELECT")) {
                ScrollableDataResult scrollableDataResult = (ScrollableDataResult) dataResult
                        .getDataObject();
                result = scrollableDataResult.getSourceBean();
            } // if (type.equalsIgnoreCase("SELECT"))
            else {
                InformationDataResult informationDataResult = (InformationDataResult) dataResult
                        .getDataObject();
                result = new Boolean(informationDataResult.getAffectedRows() > 0);
            } // if (type.equalsIgnoreCase("SELECT")) else
            return result;
        } // try
        catch (Exception ex) {
            TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.CRITICAL,
                    "QueryExecutor::executeQuery:", ex);
            throw ex;
        } // catch (Exception ex) try
        finally {
            Utils.releaseResources(null, sqlCommand, dataResult);
        }
    }

    /**
     * Trattamento dei parametri "standard"
     * 
     * @param requestContainer
     * @param responseContainer
     * @param parameter
     * @param inputParameters
     * @param dataConnection
     * @return
     */
    protected static boolean handleParameter(final RequestContainer requestContainer,
            final ResponseContainer responseContainer, SourceBean parameter,
            ArrayList inputParameters, DataConnection dataConnection) {

        return handleParameter(requestContainer, responseContainer, parameter, inputParameters,
                dataConnection, false, null, null);
    }

    /**
     * Trattamento dei parametri di tipo FILTER
     * 
     * @param requestContainer
     * @param responseContainer
     * @param parameter
     * @param inputParameters
     * @param dataConnection
     * @param statement
     * @param condizioneSql
     * @return
     */
    protected static boolean handleFilterParameter(final RequestContainer requestContainer,
            final ResponseContainer responseContainer, SourceBean parameter,
            ArrayList inputParameters, DataConnection dataConnection, final StringBuffer statement,
            final String condizioneSql) {

        return handleParameter(requestContainer, responseContainer, parameter, inputParameters,
                dataConnection, true, statement, condizioneSql);
    }

    /**
     * Gestione generalizzata per i parametri di tipo "normale" e FILTER
     * 
     * @param requestContainer
     * @param responseContainer
     * @param parameter
     * @param inputParameters
     * @param dataConnection
     * @param isFilterParameter
     * @param statement
     * @param condizioneSql
     * @return
     */
    protected static boolean handleParameter(final RequestContainer requestContainer,
            final ResponseContainer responseContainer, final SourceBean parameter,
            ArrayList inputParameters, DataConnection dataConnection,
            final boolean isFilterParameter, StringBuffer statement, final String condizioneSql) {

        boolean parameterUsed = false;
        // Lettura attributi di "tipo" del parametro
        String parameterType = (String) parameter.getAttribute("TYPE");
        String parameterValue = (String) parameter.getAttribute("VALUE");
        String parameterScope = (String) parameter.getAttribute("SCOPE");
        String parameterObjectType = (String) parameter.getAttribute("PARTYPE");

        String inParameterValue = null;
        if (parameterType.equalsIgnoreCase("ABSOLUTE"))
            inParameterValue = parameterValue;
        else {
            Object parameterValueObject = ContextScooping.getScopedParameter(requestContainer,
                    responseContainer, parameterValue, parameterScope, parameter);
            if (parameterValueObject != null)
                inParameterValue = parameterValueObject.toString();
        } // if (parameterType.equalsIgnoreCase("ABSOLUTE")) else
        if (inParameterValue == null)
            inParameterValue = "";

        if (!isFilterParameter) { // In caso di parametro "normale"
        	if (parameterObjectType == null)  {
                inputParameters
                .add(dataConnection.createDataField("", Types.VARCHAR, inParameterValue));
        	} else if (parameterObjectType.equalsIgnoreCase("INTEGER")) {
                inputParameters
                .add(dataConnection.createDataField("", Types.INTEGER, inParameterValue));
        	} 
        	else if (parameterObjectType.equalsIgnoreCase("DOUBLE")) {
                inputParameters
                .add(dataConnection.createDataField("", Types.DOUBLE, inParameterValue));
        	} 
        	else {
                inputParameters
                .add(dataConnection.createDataField("", Types.VARCHAR, inParameterValue));
        	}
            parameterUsed = true;
        } else { // In caso di parametro di filtro

            // Se il parametro di filtro HA VERAMENTE UN VALORE,
            if (!inParameterValue.equals("")) {

                // Lettura attributo di "sql" del parametro
                String sqlToAdd = (String) parameter.getAttribute("SQL");
                int indexLike = sqlToAdd.indexOf(" LIKE ");

                // Aggiungo il parametro tra i parametri di input dell'SQL, se
                // necessario
                int countParams = StringUtils.count(sqlToAdd, '?');
                if (countParams != 0) {
                    // Aggiungo il parametro tante volte quanti sono i caratteri
                    // '?' presenti nella stringa
                    for (int i = 0; i < countParams; i++) {
                        inputParameters.add(dataConnection.createDataField("", Types.VARCHAR,
                                inParameterValue + (indexLike >= 0 ? "%" : "")));
                    }
                    statement.append(condizioneSql);
                }

                // E aggiungo la parte di SQL allo statement (modifico
                // direttamente l'oggetto StringBuffer)
                statement.append(sqlToAdd);

                parameterUsed = true;

            } // if !inParameterValue
        } // else !isFilterParameter
        return parameterUsed;
    } // elaboraParametro

    /**
     * Gestione dell'opzione ORDER_BY
     * 
     * @param parameter
     * @param statement
     * @param condizioneSql
     */
    protected static void handleOrderByParameter(final SourceBean parameter,
            StringBuffer statement, final String condizioneSql) {

        // Lettura attributo di "sql" del parametro
        String sqlToAdd = (String) parameter.getAttribute("SQL");

        // Aggiungo la parte di SQL allo statement (modifico direttamente
        // l'oggetto StringBuffer)
        statement.append(condizioneSql);
        statement.append(sqlToAdd);

    } // elaboraParametroOrderBy

} // public abstract class QueryExecutor
